/*
 * 
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 * 
 * Copyright 2007-2008 Sun Microsystems, Inc. All rights reserved.
 * 
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common Development
 * and Distribution License("CDDL") (collectively, the "License").  You
 * may not use this file except in compliance with the License. You can obtain
 * a copy of the License at https://glassfish.dev.java.net/public/CDDL+GPL.html
 * or glassfish/bootstrap/legal/LICENSE.txt.  See the License for the specific
 * language governing permissions and limitations under the License.
 * 
 * When distributing the software, include this License Header Notice in each
 * file and include the License file at glassfish/bootstrap/legal/LICENSE.txt.
 * Sun designates this particular file as subject to the "Classpath" exception
 * as provided by Sun in the GPL Version 2 section of the License file that
 * accompanied this code.  If applicable, add the following below the License
 * Header, with the fields enclosed by brackets [] replaced by your own
 * identifying information: "Portions Copyrighted [year]
 * [name of copyright owner]"
 * 
 * Contributor(s):
 * 
 * If you wish your version of this file to be governed by only the CDDL or
 * only the GPL Version 2, indicate your decision by adding "[Contributor]
 * elects to include this software in this distribution under the [CDDL or GPL
 * Version 2] license."  If you don't indicate a single choice of license, a
 * recipient has the option to distribute your version of this file under
 * either the CDDL, the GPL Version 2 or to extend the choice of license to
 * its licensees as provided above.  However, if you add GPL Version 2 code
 * and therefore, elected the GPL Version 2 license, then the option applies
 * only if the new code is made subject to such option by the copyright
 * holder.
 *
 */

package com.sun.grizzly.http;

import com.sun.grizzly.DefaultSelectionKeyHandler;
import com.sun.grizzly.tcp.Response;
import com.sun.grizzly.tcp.Response.ResponseAttachment;
import com.sun.grizzly.util.Copyable;
import com.sun.grizzly.util.SelectionKeyAttachment;
import com.sun.grizzly.util.ThreadAttachment;
import java.nio.channels.SelectionKey;
import java.util.Iterator;
import java.util.logging.Level;

/**
 * Default HTTP SelectionKeyHandler implementation
 * 
 * @author Jean-Francois Arcand
 * @author Alexey Stashok
 */
public class SelectorThreadKeyHandler extends DefaultSelectionKeyHandler {
       
    private SelectorThread selectorThread;
    
    private boolean disableOptimization = false;

    public SelectorThreadKeyHandler() {
    }

    public SelectorThreadKeyHandler(SelectorThread selectorThread) {
        this.selectorThread = selectorThread;
    }

    @Override
    public void copyTo(Copyable copy) {
        super.copyTo(copy);
        SelectorThreadKeyHandler copyHandler = (SelectorThreadKeyHandler) copy;
        copyHandler.selectorThread = selectorThread;
    }

    @Override
    public void doRegisterKey(SelectionKey key, int ops, long currentTime) {
        if (!key.isValid() || !selectorThread.getKeepAliveCounter().trap(key)) {
            selectorThread.cancelKey(key);
            return;
        }
        
        // If the SelectionKey is used for continuation, do not allow
        // the key to be registered.
          if (selectorThread.getEnableAsyncExecution() && key.attachment() instanceof String){
              key.attach(null); 
              return;
          }

        if (key.isValid()){
            key.interestOps(key.interestOps() | ops);
            Object attachment = key.attachment();
            // By default, attachment a null.
            if (attachment == null) {
                key.attach(currentTime);
            } else if (attachment instanceof ThreadAttachment) {
                ((ThreadAttachment) attachment).setTimeout(currentTime);
            }
        }//todo should we not cancel the key here ?
    }

    @Override
    public void cancel(SelectionKey key) {
        if (key != null) {
            selectorThread.getKeepAliveCounter().untrap(key);
        }
        if (selectorThread.isMonitoringEnabled()) {
            selectorThread.getRequestGroupInfo().decreaseCountOpenConnections();
        }
        super.cancel(key);
    }
    
    
    /**
     * Add a {@link SelectionKey} to the banned list of SelectionKeys. 
     * A SelectionKey is banned when new registration aren't allowed on the 
     * Selector.
     */
    public void addBannedSelectionKey(SelectionKey key){
        key.attach("banned");
    }
    
    /**
     * Reset the expiration time
     */
    public void resetExpiration(){
        disableOptimization = true;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void expire(Iterator<SelectionKey> iterator) {
        if (timeout <= 0) return;
        
        long currentTime = System.currentTimeMillis();
        if (!disableOptimization && currentTime < nextKeysExpiration) {
            return;
        }
        nextKeysExpiration = currentTime + timeout;        
                        
        SelectionKey key;
        while (iterator.hasNext()) {
            key = iterator.next();
            
            if (!key.isValid()){
                continue;
            }

            Long expire = getExpirationStamp(key);
            if (expire != null){
                if (currentTime - expire >= timeout) {                    
                    Object attachment = key.attachment();
                    if (attachment != null) {
                        if (attachment instanceof Response.ResponseAttachment) {
                            ((ResponseAttachment) attachment).timeout();                            
                            key.attach(null);
                        } else {
                            cancel(key);
                        }
                    } else {
                        cancel(key);
                    }
                } else if (expire + timeout < nextKeysExpiration) {
                    nextKeysExpiration = expire + timeout;
                }
            }
        }
    }

   
    
    /**
     * Gets expiration timeout stamp from the {@link SelectionKey} 
     * depending on its attachment
     * 
     * @param {@link SelectionKey}
     */
    private Long getExpirationStamp(SelectionKey key) {
        Object attachment = key.attachment();
        if (attachment != null) {
            try {                            
                if (attachment instanceof Long) {
                    return (Long) attachment;
                } else if (attachment instanceof SelectionKeyAttachment) {
                    return ((SelectionKeyAttachment) attachment).getTimeout();
                } else if (attachment instanceof Response.ResponseAttachment) {
                    return ((ResponseAttachment) attachment).getExpirationTime()
                            - timeout;
                }
            } catch (ClassCastException ex) {
                if (logger.isLoggable(Level.FINEST)) {
                    logger.log(Level.FINEST, 
                            "Invalid SelectionKey attachment", ex);
                }
            }
        }

        return null;
    }
    
}